from keras.callbacks import EarlyStopping, ModelCheckpoint, Callback
import tensorflow as tf
import keras.backend as K
import pandas as pd
from keras.models import Sequential, Model
from keras.layers import CuDNNLSTM, Dense, Conv1D, Conv2D, Reshape, TimeDistributed, Flatten, Dropout, Input, Embedding, Concatenate, Activation, BatchNormalization, Permute, GlobalAveragePooling1D
from keras.layers.merge import concatenate, multiply
from tcn import compiled_tcn, TCN
import matplotlib.pyplot as plt
import tensorflow as tf
from sklearn.metrics import roc_auc_score, f1_score
from keras.optimizers import Adam
from keras.regularizers import l2
import numpy as np
from metrics import f1_actual

adam = lambda lr: Adam(lr=lr)
reg = lambda l2reg: l2(l2reg)


def rebalance_dfs(dfs, label_col, seed):
    balance = 0
    pos_samples = {}
    for i, df in enumerate(dfs):
        pos_samples[i] = df.loc[df[label_col]==1].shape[0]
        balance = max([balance, pos_samples[i]/len(df)])
    to_return = []
    for i, df in enumerate(dfs):
        samples = int(pos_samples[i]/balance)
        neg_samples = samples-pos_samples[i]
        df = pd.concat([df.loc[df[label_col]==1],df.loc[df[label_col]==0].sample(n=neg_samples, random_state=seed)])
        df = df.sort_values(["Turbine", "TimeStamp"])
        to_return.append(df)
    return to_return

def auc(y_true, y_pred):
    auc = tf.metrics.auc(y_true, y_pred)[1]
    K.get_session().run(tf.local_variables_initializer())
    return auc

class roc_auc_callback(Callback):
    def __init__(self,tr_gen,y_train, val_gen, y_val):
        self.tr_gen = tr_gen
        self.y = y_train
        self.val_gen = val_gen
        self.y_val = y_val

    def on_train_begin(self, logs={}):
        return

    def on_train_end(self, logs={}):
        return

    def on_epoch_begin(self, epoch, logs={}):
        return

    def on_epoch_end(self, epoch, logs={}):
        y_pred = self.model.predict_generator(self.tr_gen, verbose=0)
        roc = roc_auc_score(self.y, y_pred)
        logs['roc_auc'] = roc # roc_auc_score(self.y, y_pred)

        y_pred_val = self.model.predict_generator(self.val_gen, verbose=0)
        roc_val = roc_auc_score(self.y_val, y_pred_val)
        logs['roc_auc_val'] = roc_val #roc_auc_score(self.y_val, y_pred_val)

        print('\rroc_auc: %s - roc_auc_val: %s ' % (str(round(roc,5)),str(round(roc_val,5))), end=10*' '+'\n')
        return

    def on_batch_begin(self, batch, logs={}):
        return

    def on_batch_end(self, batch, logs={}):
        return

class f1_callback(Callback):
    def __init__(self,tr_gen,y_train, val_gen, y_val, seq_len):
        self.tr_gen = tr_gen
        self.y = y_train
        self.val_gen = val_gen
        self.y_val = y_val
        self.win = seq_len

    def on_train_begin(self, logs={}):
        return

    def on_train_end(self, logs={}):
        return

    def on_epoch_begin(self, epoch, logs={}):
        return

    def on_epoch_end(self, epoch, logs={}):
        y_pred = self.model.predict_generator(self.tr_gen, verbose=0)
        f1 = f1_actual(self.y, y_pred, self.win)
        logs['f1'] = f1 #f1_score(self.y, y_pred)

        y_pred_val = self.model.predict_generator(self.val_gen, verbose=0)
        f1_val = f1_actual(self.y_val, y_pred_val, self.win)
        logs['f1_val'] = f1_val #roc_auc_score(self.y_val, y_pred_val)

        print('\rf1: %s f1_val: %s ' % (str(round(f1,5)), str(round(f1_val,5))), end=10*' '+'\n')
        return

    def on_batch_begin(self, batch, logs={}):
        return

    def on_batch_end(self, batch, logs={}):
        return

def auc_2(y_true, y_pred):
    return tf.py_func(roc_auc_score, (y_true, y_pred), tf.double)


def get_callbacks_OLD(model_save_dir, EARLY_STOPPING_PATIENCE, tr_gen, y_tr, val_gen, y_val, save=False):
    cb = [roc_auc_callback(tr_gen, y_tr, val_gen, y_val), EarlyStopping(monitor='roc_auc_val', mode="max", patience=EARLY_STOPPING_PATIENCE)]
    if save:
        cb.append(ModelCheckpoint(filepath=model_save_dir, monitor='roc_auc_val', mode="max", save_best_only=True, verbose=1))
    return cb


def get_callbacks(model_save_dir, EARLY_STOPPING_PATIENCE, tr_gen, y_tr, val_gen, y_val, seq_len, save=False):
    cb = [roc_auc_callback(tr_gen, y_tr, val_gen, y_val), f1_callback(tr_gen, y_tr, val_gen, y_val, seq_len), EarlyStopping(monitor='f1_val', mode="max", patience=EARLY_STOPPING_PATIENCE)]
    if save:
        cb.append(ModelCheckpoint(filepath=model_save_dir, monitor='f1_val', mode="max", save_best_only=True, verbose=1))
    return cb

def get_emb(n_unique, seq_len):
    emb_size = min(n_unique//2, 50)
    inp = Input(shape=(1,seq_len))
    emb = Embedding(n_unique, emb_size, input_length=seq_len)(inp)
    emb = Reshape(target_shape=(emb_size,seq_len))(emb)
    return inp, emb

def build_embedding_network(emb_cols, num_cols, seq_len):
    # emb_cols: dict of col_no: n_unique
    inputs = []
    embeddings = []
    for col_no, c in emb_cols.items():
        inp, emb = get_emb(c, seq_len)
        inputs.append(inp)
        embeddings.append(emb)
    
    num_numeric = len(num_cols)
    input_numeric = Input(shape=(num_numeric,seq_len))
    inputs.append(input_numeric)
    embeddings.append(input_numeric)
    out = Concatenate(axis=1)(embeddings)
    out = Reshape(target_shape=(seq_len,-1))(out)
    return inputs, out

def get_tcn_model(emb_cols, num_cols, seq_len, lr, dropout_frac, w_reg, a_reg):
    #i = Input(batch_shape=(batch_size, seq_len, n_features))
    inp, out = build_embedding_network(emb_cols, num_cols, seq_len)
    #o = Dropout(0.2, noise_shape=(batch_size,1,n_features))(i)
    o = TCN(dropout_rate=0.3, nb_filters=32, nb_stacks=2, return_sequences=False)(out)  # The TCN layers are here.
    #o = Flatten()(o)
    o = Dense(128, activation="relu")(o)
    o = Dropout(0.5)(o)
    #o = Flatten()(o)
    o = Dense(1, activation="sigmoid")(o)
    m = Model(inputs=inp, outputs=[o])
    m.compile(optimizer=adam, loss='binary_crossentropy', metrics=[])
    return m

def get_mlp_model(emb_cols, num_cols, seq_len, lr, dropout_frac, w_reg, a_reg):
    inp, out = build_embedding_network(emb_cols, num_cols, seq_len)

    z = Flatten()(out)
    z = Dense(32, kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg))(z)
    z = BatchNormalization()(z)
    z = Activation('relu')(z)
    z = Dropout(dropout_frac)(z)
    z = Dense(16, kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg))(z)
    z = BatchNormalization()(z)
    z = Activation('relu')(z)
    z = Dropout(dropout_frac)(z)
    z = Dense(1, activation='sigmoid')(z)

    model = Model(inp, z, name="mlp")
    model.compile(optimizer=adam(lr), loss='binary_crossentropy', metrics=[])
    return model


def get_cnn_model(emb_cols, num_cols, seq_len, lr, dropout_frac, w_reg, a_reg):
    inp, out = build_embedding_network(emb_cols, num_cols, seq_len)
    
    y = Conv1D(activation="relu", filters=32, kernel_size=3, kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg))(out)
    y = BatchNormalization()(y)
    y = Activation('relu')(y)
    y = Dropout(dropout_frac)(y)
    y = squeeze_excite_block(y)
    y = TimeDistributed(Dense(8, activation="relu", kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg)))(y)
    y = Dropout(dropout_frac)(y)
    y = Flatten()(y)

    y = Dense(1, activation="sigmoid")(y)

    model = Model(inp, y, name="cnn")
    model.compile(optimizer=adam(lr), loss='binary_crossentropy', metrics=[])
    return model


def get_lstm_model(emb_cols, num_cols, seq_len, lr, dropout_frac, w_reg, a_reg):
    inp, out = build_embedding_network(emb_cols, num_cols, seq_len)
    
    x = CuDNNLSTM(32, return_sequences=True)(out)
    x = BatchNormalization()(x)
    x = Dropout(dropout_frac)(x)
    x = CuDNNLSTM(4, return_sequences=True)(x)
    x = BatchNormalization()(x)
    x = Dropout(dropout_frac)(x)
    x = TimeDistributed(Dense(8, activation="relu", kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg)))(x)
    x = Flatten()(x)

    x = Dense(1, activation='sigmoid')(x)

    model = Model(inp, x, name="lstm")
    model.compile(optimizer=adam(lr), loss='binary_crossentropy', metrics=[])
    return model

def get_hmvts_model(emb_cols, num_cols, seq_len, lr, dropout_frac, w_reg, a_reg):
    inp, out = build_embedding_network(emb_cols, num_cols, seq_len)

    x = CuDNNLSTM(32, return_sequences=True)(out)
    x = BatchNormalization()(x)
    x = Dropout(dropout_frac)(x)
    x = CuDNNLSTM(4, return_sequences=True)(x)
    x = BatchNormalization()(x)
    x = Dropout(dropout_frac)(x)
    x = TimeDistributed(Dense(8, activation="relu", kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg)))(x)
    x = Flatten()(x)


    y = Conv1D(activation="relu", filters=32, kernel_size=3, kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg))(out)
    y = BatchNormalization()(y)
    y = Activation('relu')(y)
    y = Dropout(dropout_frac)(y)
    y = squeeze_excite_block(y)

    y = TimeDistributed(Dense(8, activation="relu", kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg)))(y)
    y = Dropout(dropout_frac)(y)
    y = Flatten()(y)

    z = Flatten()(out)
    z = Dense(32, kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg))(z)
    z = BatchNormalization()(z)
    z = Activation('relu')(z)
    z = Dropout(dropout_frac)(z)
    z = Dense(8, kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg))(z)
    z = BatchNormalization()(z)
    z = Activation('relu')(z)
    z = Dropout(dropout_frac)(z)

    o = concatenate([x, y, z])
    o = Dense(4, kernel_regularizer=reg(w_reg), activity_regularizer=reg(a_reg))(o)
    o = Dropout(dropout_frac)(o)

    output = Dense(1, activation='sigmoid')(o)

    model = Model(inp, output, name="hmvts")
    model.compile(optimizer=adam(lr), loss="binary_crossentropy", metrics=[])
    return model

def save_history_plot(hist, savedir):
    plt.clf()
    # Get training and test loss histories
    training_loss = hist.history['loss']
    #training_auc = hist.history['roc_auc']
    val_loss = hist.history['val_loss']
    #val_auc = hist.history['roc_auc_val']
    # Create count of the number of epochs
    epoch_count = range(1, len(training_loss) + 1)

    # Visualize loss history
    plt.plot(epoch_count, training_loss, 'r--')
    plt.plot(epoch_count, val_loss, 'b--')
    #plt.plot(epoch_count, training_auc, 'r-')
    #plt.plot(epoch_count, val_auc, 'b-')
    plt.legend(['Training Loss', 'Validation Loss'])
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.savefig(savedir)
    plt.clf()
    return #max(training_auc), max(val_auc)

def squeeze_excite_block(input):
    ''' Create a squeeze-excite block
    Args:
        input: input tensor
        filters: number of output filters
        k: width factor
    Returns: a keras tensor
    '''
    filters = input._keras_shape[-1] # channel_axis = -1 for TF

    se = GlobalAveragePooling1D()(input)
    se = Reshape((1, filters))(se)
    se = Dense(filters // 16,  activation='relu', kernel_initializer='he_normal', use_bias=False)(se)
    se = Dense(filters, activation='sigmoid', kernel_initializer='he_normal', use_bias=False)(se)
    se = multiply([input, se])
    return se


def f1(y_true, y_pred):
    y_pred = K.round(y_pred)
    tp = K.sum(K.cast(y_true*y_pred, 'float'), axis=0)
    tn = K.sum(K.cast((1-y_true)*(1-y_pred), 'float'), axis=0)
    fp = K.sum(K.cast((1-y_true)*y_pred, 'float'), axis=0)
    fn = K.sum(K.cast(y_true*(1-y_pred), 'float'), axis=0)

    p = tp / (tp + fp + K.epsilon())
    r = tp / (tp + fn + K.epsilon())

    f1 = 2*p*r / (p+r+K.epsilon())
    f1 = tf.where(tf.is_nan(f1), tf.zeros_like(f1), f1)
    return K.mean(f1)

def f1_loss(y_true, y_pred):
    
    tp = K.sum(K.cast(y_true*y_pred, 'float'), axis=0)
    tn = K.sum(K.cast((1-y_true)*(1-y_pred), 'float'), axis=0)
    fp = K.sum(K.cast((1-y_true)*y_pred, 'float'), axis=0)
    fn = K.sum(K.cast(y_true*(1-y_pred), 'float'), axis=0)

    p = tp / (tp + fp + K.epsilon())
    r = tp / (tp + fn + K.epsilon())

    f1 = 2*p*r / (p+r+K.epsilon())
    f1 = tf.where(tf.is_nan(f1), tf.zeros_like(f1), f1)
    return 1 - K.mean(f1)