import pandas as pd
import numpy as np
import keras

class TSGen(keras.utils.Sequence):
    """Utility class for generating batches of temporal data.
    """

    def __init__(self, df, targets, length,
                 start_index=0,
                 end_index=None,
                 shuffle=False,
                 reverse=False,
                 batch_size=128,
                 emb_cols=None,
                 num_cols=None,
                 group_col="Turbine",
                 seed=None):
        if len(df) != len(targets):
            raise ValueError('Data and targets have to be' +
                             ' of same length. '
                             'Data length is {}'.format(len(df)) +
                             ' while target length is {}'.format(len(targets)))
        if seed is not None:
            np.random.seed(seed)
        self.data = df.values
        self.group_col = group_col
        self.group_indices = self._get_group_indices(df.loc[:,self.group_col].copy())
        self.num_groups = len(self.group_indices)
        print("num_groups", self.num_groups)
        self.targets = targets
        self.length = length
        self.group_offset = (self.length * self.num_groups)
        self.start_index = start_index + self.group_offset
        if end_index is None:
            end_index = len(self.data) - 1
        self.end_index = end_index
        self.shuffle = shuffle
        self.reverse = reverse
        self.batch_size = batch_size
        self.emb_cols = emb_cols
        self.num_cols = num_cols 
        print("start index", self.start_index)
        print("end index", self.end_index)
        print("length", len(self))
        print("data shape", self.data.shape)
        print("target shape", self.targets.shape)
       
    def _get_group_indices(self, series):
        series = series.reset_index(drop=True)
        new_group_series = series.diff()
        split_indices = new_group_series[new_group_series==1].index.values
        return split_indices
    
    def _get_rev_group_no(self, row):
        try:
            rev_group = np.where(row-self.start_index>self.group_indices)[0][-1]+1
        except IndexError:
            rev_group = 0
        return self.num_groups-rev_group
    
    def __len__(self):
        # Number of batches
        return (self.end_index - self.start_index - self.length) // (self.batch_size)

    def __getitem__(self, index):
        # Get one batch
        # Should return a tuple samples, targets, where samples are list of numpy arrays and targets is a numpy array.
        if self.shuffle:
            rows = np.random.randint(self.start_index, self.end_index + 1 - self.length, size=self.batch_size)
        else:
            i = self.start_index + (self.batch_size * index)
            rows = np.arange(i, min(i + self.batch_size, self.end_index + 1))
        
        row_group = {row:self._get_rev_group_no(row) for row in rows}
        
        actual_inds = np.array([row - (self.length*row_group[row]) for row in rows])
        #print(len(actual_inds))
        #print("actual inds", actual_inds)
        assert actual_inds.max()+self.length <= self.end_index
        
        samples = []
        for e in self.emb_cols:
            emb_samples = np.array([self.data[ind:ind + self.length, e] for ind in actual_inds])
            emb_samples = emb_samples[:, np.newaxis]
            #print(emb_samples.shape)
            samples.append(emb_samples)
        #print("rows", rows)
        #print("row groups", row_group)
        #print("index:",index)
        num_samples =  np.array([self.data[ind:ind + self.length, self.num_cols] for ind in actual_inds])
        #print(num_samples.shape)
        num_reshaped = np.reshape(num_samples, (len(actual_inds), len(self.num_cols), self.length))
        #print(num_reshaped.shape)
        samples.append(num_reshaped)
        targets = np.array([self.targets[ind+self.length] for ind in actual_inds])
        return samples, targets